// File       : UAdaptEM.cc
// Creation   : Feb 97
// Creator    : N.OLIVER
// Goal       : TAdaptEM class definition. See "UAdaptEM.h".
//
// Tab        : yes


// unit include
#include "UAdaptEM.h"
#include "matmath.h"
#include "qmGaussian.h++"

// project includes
#include "UObjAlloc.h"

// ANSI includes
#include <math.h>
#include <bstring.h>

// system includes
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <termios.h>
#include <stdlib.h>
#include <stdio.h>

////////// defines
#define SIEVEBOUND     1000.
#define INITVARWEIGHT  10.  
#define MAXEXP         100.
#define MIN(a,b)       (((a)<(b))?(a):(b))
#define MAX(a,b)       (((a)<(b))?(b):(a))
#define MIN_FLOAT      .0001

////////// static data members initialisation


//
//    Init
//

ErrorNumber TAdaptEM::Init ( THistogram      *histo,
			     TBitMap         *source,
	                     TLearnIndex     *ind,
			     TROIMutex       *roi,
			     UInt32           maxNumComp,
			     TGenGaussian    *models,
			     UInt32           numComp,
			     float           *weights,
			     UInt16           dim,
			     double           initVar,
			     float            updateTime,
			     double           Tc
			     )
{
  ErrorNumber       err = kErrNoErr;
  int               i;
  // init data members
  kMaxTerms = maxNumComp;
  // default maximum number of terms
  fNumTerms = 0;
  fTimer = (TTimer*)NULL;

  fw = new double[kMaxTerms];
  kInitVar = initVar;
  fModels = new TGenGaussian[kMaxTerms];
  fGaussOutputs = new double[kMaxTerms];
  fScaledOutputs = new double[kMaxTerms];
  fAverageDist = new double[kMaxTerms];
  fUpdateTime = updateTime;
  kNumIter = 1;      // only one iteration
  kMaxObs = 15000;   // maximum number of observations

  kTc = Tc;

  // Initialize data members
  bzero(fw, sizeof(double)*kMaxTerms);
  bzero(fScaledOutputs, sizeof(double)*kMaxTerms);
  bzero(fGaussOutputs, sizeof(double)*kMaxTerms);
  bzero(fAverageDist, sizeof(double)*kMaxTerms);
  
  fInputData = (double**)malloc( sizeof(double*)*kMaxObs );
  for ( i=0;i<kMaxObs; i++)
  {
    fInputData[i] = new double[dim];
  }

  for (i=0;i<kMaxTerms;i++)
  fModels[i].Init(dim);
  
  if ( models != NULL )
  {
    fNumTerms = numComp;
    for (i=0;i<fNumTerms;i++)
    {
      fModels[i].set(models[i]);
      if ( weights != NULL )
      {
	fw[i] = (double)weights[i];
      } else
      {
	fw[i] = (double)1./(double)fNumTerms;
      }
      cout << "initial fw[i] " << fw[i] << endl;
    }
  }
  err = AllocAndInit (fTimer);
  // error handling
  if (err != kErrNoErr)
  this->Dispose ();

  fTimer->Gap();

  // call inherited::Init
  err = TLearn::Init ( histo, source, ind, roi );

  // error handling
  if (err != kErrNoErr)
  this->Dispose ();
  
  return err;
}




//
//    Dispose
//

void TAdaptEM::Dispose ()
{
  // call inherited::Dispose
  TLearn::Dispose ();

  // Dispose data members
  delete fw;
  delete fModels;
  delete fAverageDist;
  delete fGaussOutputs;
  delete fScaledOutputs;
  for ( int i=0; i<kMaxObs; i++ )
  {
    delete fInputData[i];
  }
  delete fInputData;
  // dispose the timer
  TestDisposeAndFree (fTimer);
}


//
//    Reads data and saves them in fInputData
//

ErrorNumber
TAdaptEM::ReadData()
{
  UInt32         offset;
  UInt8         *basePtr, *linePtr, *pixPtr;       // bitmap pointers
  UInt32         r, g, b, lum;                     // RGB components
  int            i, j,width, height;
  ErrorNumber    err = kErrNoErr;

  fNumObs = 0;
  // Read data points and include them in the model
  err =  InitProcess( width, height, offset, basePtr );

  if (err != kErrNoErr)
  return err;
  
  for ( i=0; i<height/2; i++ ) 
  {
    for ( j=0; j<width/2;j++ )
    {
      pixPtr=basePtr+i*offset+j*4;
      pixPtr++;
      b = *(pixPtr);
      pixPtr++;
      g = *pixPtr;
      pixPtr++;
      r = *pixPtr;
      pixPtr++;
      // compute pixel luminance
      lum = r+g+b;
      fInputData[fNumObs][0] = ((lum > 0 ) ? (double)r/(double)lum : 0);
      fInputData[fNumObs][1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
      fNumObs++;
      if ( j>0 )
	{
	  pixPtr=basePtr+i*offset-j*4;
	  pixPtr++;
	  b = *pixPtr;
	  pixPtr++;
	  g = *pixPtr;
	  pixPtr++;
	  r = *pixPtr;
	  pixPtr++;
	  // compute pixel luminance
	  lum = r+g+b;
	  fInputData[fNumObs][0] = ((lum > 0 ) ? (double)r/(double)lum : 0);
	  fInputData[fNumObs][1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
	  fNumObs++;
	  if ( i>0 )
	    {
	      pixPtr=basePtr-i*offset-j*4;
	      pixPtr++;
	      b = *pixPtr;
	      pixPtr++;
	      g = *pixPtr;
	      pixPtr++;
	      r = *pixPtr;
	      pixPtr++;
	      // compute pixel luminance
	      lum = r+g+b;
	      fInputData[fNumObs][0] = ((lum > 0 ) ? (double)r/(double)lum : 0);
	      fInputData[fNumObs][1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
	      fNumObs++;
	    }
	} else if ( i>0 )
	  {
	    pixPtr=basePtr-i*offset+j*4;
	    pixPtr++;
	    b = *pixPtr;
	    pixPtr++;
	    g = *pixPtr;
	    pixPtr++;
	    r = *pixPtr;
	    pixPtr++;
	    // compute pixel luminance
	    lum = r+g+b;
	    fInputData[fNumObs][0] = ((lum > 0 ) ? (double)r/(double)lum : 0);
	    fInputData[fNumObs][1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
	    fNumObs++;
	  }
    } // for j
  } // for i      
  cout << "Num Obs " << fNumObs;
  return err;
}


//
//     GetGaussianOutputs
//
//   Get the values for the individual terms.  
//   The output is not updated by this routine.  
//   No learning (updating of terms) is done here either.
//   Get the value of every Gaussian term at fInput[] 
//   and store scaled and unscaled values.

ErrorNumber
TAdaptEM::GetGaussianOutputs()
{
  ErrorNumber      err  = kErrNoErr;  
  int              i,j,k;
  int              dim = fModels[0].getDim();
  double           mahal2;
  double*          m;
  double*          ic;

  fOutput = 0.0;
  for(i=0;i<fNumTerms;i++){
      if(fw[i] > 0.0)
      {
	mahal2=0.0;
	for(j=0;j<dim;j++)
	{
	  for(k=0;k<dim;k++)
	  {
	    m=fModels[i].getMean();
	    ic=fModels[i].getInvCov(); 
	    mahal2 += (m[j] - fInput[j])*(m[k] - fInput[k])*
					 ic[j*dim+k];
	  } // for k
	} // for j

	if(mahal2 > MAXEXP)
	fGaussOutputs[i] = 0.0;
	else
	fGaussOutputs[i] = exp(-.5*mahal2);

	fScaledOutputs[i] = 
			fGaussOutputs[i]/(pow(2.*M_PI,dim/2.)*sqrt(fModels[i].getDet()));
	fOutput += fw[i]*fScaledOutputs[i];
      } else
      {
	fOutput = fScaledOutputs[i] = fGaussOutputs[i] = 0.0;
      }
  }
  
  return(err);
}

//
//      AddTerm
//

ErrorNumber
TAdaptEM::AddTerm()
/* Create a new term.                   */
/* This is C() in Equation (4.6),       */
/* defined by Equations (3.12) - (3.16) */
{
  ErrorNumber       err = kErrNoErr;
  int               i,j,k;
  int               dim = fModels[0].getDim();
  double            tmp;
  double            tvar=0.0;
  double            *tcov;
  double            totalscaled=0.0;
  int               usetvar=0;
  double            *auxP, *auxP1;

  if(fNumTerms >= kMaxTerms)
  {
    fprintf(stderr,"\n%d term",fNumTerms);
    fprintf(stderr,"Overflow:  kMaxTerms.\n");
    fprintf(stderr,"Cannot create new term.\n");
    exit(1);
  } 

/* Determine tvar from existing term. */

/* find total output of all terms */
  for(i=0;i<fNumTerms;i++) totalscaled+=fScaledOutputs[i];

/* if total output is zero, then use default */
  if(totalscaled==0.0 || fNumTerms == 0) 
  {
    tvar = kInitVar;
    usetvar = -1;
  }
/* else make the variance the weighted average 
   of all other terms */
  else 
  {
    usetvar = 0;
    tcov = new double[ dim*dim ];
    for(i=0;i<fNumTerms;i++) 
    {
      for(j=0;j<dim;j++) 
      {
	for(k=0;k<dim;k++) 
	{
	  tcov[j*dim+k] += (fScaledOutputs[i]/totalscaled)*fModels[i].getElemCov(j*dim+k);
	}
      }
    } // for i
  } // else

/* Now set up the new term. */
  
  for(i=0;i<dim;i++)
  {
    fModels[fNumTerms].setElemMean( fInput[i], i);
    if(usetvar)
    {
      fModels[fNumTerms].setElemCov( tvar, i*dim+i );
    }
    else
    {
      for(j=0;j<dim;j++)
      fModels[fNumTerms].setElemCov( tcov[i*dim+j], i*dim+j );
    }
  } // for i
  if (fNumPts>0)
    fw[fNumTerms] = 1.0/fNumPts;
  auxP = fModels[fNumTerms].getCov();
  auxP1 = fModels[fNumTerms].getInvCov();
  if (dim==2)
  {
    DMatInvert2(auxP,&tmp,auxP1);
  } else
  {
    DMatInvert(auxP,dim,&tmp,auxP1);
  }
  fModels[fNumTerms].setDet( tmp );

  fNumTerms++;

  if(!usetvar) delete(tcov);
  return(err);
}

//
//     UpdateW
//
ErrorNumber
TAdaptEM::UpdateW(int    term,
		 double amt)
/* Update the mixture percentage of the term */
{
  ErrorNumber       err = kErrNoErr;
  if (fNumPts>1)
  fw[term] += (amt - fw[term])/(fNumPts-1.);
  return(err);
}
/*****************************************************/

/*****************************************************/
ErrorNumber
TAdaptEM::UpdateMean( int    term,
		      double amt)
/* Update the mean of the term */
{
  ErrorNumber      err = kErrNoErr;
  int              i;
  int              dim = fModels[term].getDim();
  double          *aux = fModels[term].getMean();

  if ((fNumPts>1) && fw[term]>0)
    {
      for(i=0;i<dim;i++, aux++)
	{
	  (*aux) += (amt/((fNumPts-1.)*fw[term])) *
	    (fInput[i] - (*aux));
	}
    }
  return(err);
}

/*****************************************************/

/*****************************************************/
ErrorNumber
TAdaptEM::UpdateCov(int    term,
		    double amt)
/* Update the var of the term. */
{
  ErrorNumber      err = kErrNoErr;
  int              i,j;
  int              dim = fModels[term].getDim();
  double           tmp, *auxP, *auxP1;
  double*          ptrCov = fModels[term].getCov();

  if ( (fNumPts>1) && (fw[term]>0) )
    {
      for(i=0;i<dim;i++)
	{
	  for(j=0;j<dim;j++, ptrCov++)
	    {
	      (*ptrCov) += 
		(amt/((fNumPts-1.)*fw[term]+INITVARWEIGHT)) *
		( (fModels[term].getElemMean(i) - fInput[i]) *
		  (fModels[term].getElemMean(j) - fInput[j]) - (*ptrCov) );
	      auxP = fModels[term].getCov();
	      auxP1 = fModels[term].getInvCov();
	      if ( dim==2 )
		{
		  DMatInvert2(auxP,&tmp,auxP1);
		} else
		  {
		    DMatInvert(auxP,dim,&tmp,auxP1);
		  }
	      fModels[term].setDet( tmp );
	    }
	} //for
    }
  return(err);
} /* end update cov */

//
//     LearnStep
//
//     On-line learning of the next sample
//     D_new = [1-P(zeta ; D_old)] U(zeta ; D_old) + 
//              P(zeta ; D_old)  C(zeta ; D_old)

ErrorNumber
TAdaptEM::LearnStep()
{
  ErrorNumber      err  = kErrNoErr;  
  int i,j;
  int P=1;
  double amt;
  double maxv=0.0;

  if(fNumTerms == 0){
    AddTerm();
    return err;
  }
  if(fNumTerms < kMaxTerms){
    for(i=0;i<fNumTerms;i++)
      if(fGaussOutputs[i] > maxv )
	maxv = fGaussOutputs[i];
    /* create decision */
    if( kTc >= maxv )
      P=0;
  } /* end if numterms */
  else
    {
      P=0;
    }
  if (P) 
    // C()
    {
      AddTerm();
    }
  else 
    {  
      // U
      
      for(i=0;i<fNumTerms;i++){
      
	if(fOutput > 0.0) 
	  {
	    amt = fScaledOutputs[i]*fw[i]/fOutput;
	    fAverageDist[i]+=fGaussOutputs[i];
	  }
	else 
	amt = 0.0;

	UpdateW(i,amt);
	if(amt > 0.0){
	    UpdateCov(i,amt);
	    UpdateMean(i,amt);
	  } /* end if amt */

      } /* end for */
  } /* end else (not P) */
  return(err);
}

//
//     Normalize
//
ErrorNumber
TAdaptEM::Normalize()
{
  ErrorNumber      err  = kErrNoErr;
  int i,j,k;
  int dim = fModels[0].getDim();
  double tmp;
  double *ptrTmp, *auxP, *auxP1;
  double sum=0.0;
  double under[1000],over[1000];
  double tunder=0.0,tover=0.0,piprime;

  for(i=0;i<fNumTerms;i++) sum += fw[i];
  for(i=0;i<fNumTerms;i++) fw[i] = fw[i] / sum;

  /* Implement bounds dictated by sieve S4 */
  /* Equation (4.4)                        */

  /* Bounds on p eliminated.  Uncomment if desired
     piprime=1./(SIEVEBOUND*numterms);
  */
  piprime=0.0;

  for(i=0;i<fNumTerms;i++)
  {
    if(fw[i]<piprime)
    {
      under[i]=piprime-fw[i];
      tunder+=piprime-fw[i];
      over[i]=0.0;
    }
    else under[i]=0.0;

    if(fw[i]>piprime){
	over[i]=fw[i]-piprime;
	tover+=fw[i]-piprime;
	under[i]=0.0;
      }
    else over[i]=0.0;
  }
  
  for(i=0;i<fNumTerms;i++)
  {
    if(under[i]>0.0) fw[i]=piprime;
    if(over[i]>0.0) fw[i]-=tunder*(over[i]/tover);
    if(fModels[i].getDet() <= 0.0)
    {
      for(j=0;j<dim;j++)
      {
	fModels[i].setElemCov(0.0, j*dim+j );
      }
    }
    for(j=0;j<dim;j++)
    {
      ptrTmp = fModels[i].getMean()+j;
      if(*ptrTmp>SIEVEBOUND*fNumTerms) 
      {
	(*ptrTmp)=SIEVEBOUND*fNumTerms;
      }
      if((*ptrTmp)<-SIEVEBOUND*fNumTerms)
      {
	(*ptrTmp)=-SIEVEBOUND*fNumTerms;
      }
      ptrTmp = fModels[i].getCov()+j*dim;
      if(*(ptrTmp+j)>SIEVEBOUND*fNumTerms) 
      {
	for(k=0;k<dim;k++)
	{
	  *(ptrTmp+k) = 0.0;
	}
	*(ptrTmp+j) = SIEVEBOUND*fNumTerms;
      }
      if(*(ptrTmp+j)<1./(SIEVEBOUND*fNumTerms)) 
      {
	for(k=0;k<dim;k++)
	{
	  *(ptrTmp+k) = 0.0;
	}
	*(ptrTmp+j) = 1./(SIEVEBOUND*fNumTerms);
      }
    } // for j
    auxP = fModels[i].getCov();
    auxP1 = fModels[i].getInvCov();
    if ( dim == 2 )
    {
      DMatInvert2(auxP,&tmp,auxP1);
    } else
    {
      DMatInvert(auxP,dim,&tmp,auxP1);
    }
    fModels[i].setDet(tmp);
  } // for i
  
  return(err);
} // normalize

//
//     Feed Data
//

ErrorNumber
TAdaptEM::FeedData()
{
  ErrorNumber      err  = kErrNoErr;
  for ( int i=0; i<fNumObs; i++ )
  {
    fInput = fInputData[i];
    fNumPts++;
    GetGaussianOutputs();
    LearnStep();
    Normalize();
  }

  return(err);
}

//
//     Learn using adaptive mixture of gaussians
//
ErrorNumber
TAdaptEM::Learn()
{
  ErrorNumber      err  = kErrNoErr;
  int              i,j,index,validTerms=0;
  static int       timeElapsed = 0;
  static int       firstTime=1;
  float            maxDist=0;
  TGaussian       *modelsArray;
  float           *weights = new float[fNumTerms];
  float            auxMean[2];
  float            auxCov[3];
  // read the data every X seconds
  if ( ((timeElapsed/kTimerNbTiksPerSecond)>fUpdateTime) || firstTime )
  {
    firstTime=0;
    timeElapsed=0;
    if ( ReadData() == kErrNoErr )
    {
      for (i=0;i<kNumIter;i++)
      {
	FeedData();
      }
      modelsArray = new TGaussian[fNumTerms];
      for (i=0;i<fNumTerms;i++)
      {
	if ( fw[i] < MIN_FLOAT )
	  continue;
	//	fAverageDist[i]*=fw[i];
	fAverageDist[i]/=(fNumObs*kNumIter);
	maxDist+=fAverageDist[i];
	for (j=0;j<2;j++)
	  {
	  auxMean[j]=(float)fModels[i].getElemMean(j);
	  auxCov[j]=(float)fModels[i].getElemCov(j);
	  }
	auxCov[2]=(float)fModels[i].getElemCov(3);
	
	modelsArray[validTerms].set(auxMean,
				    auxCov);

	cout << "Model " << i << ": " << endl;
	fModels[i].print(cout);
	cout << modelsArray[validTerms] << endl;
	weights[validTerms]=(float)fw[i];
	cout << "Average Distance " << fAverageDist[i] << endl;
	cout << "Weight " << weights[validTerms] << endl;
	validTerms++;

      }
      maxDist/=validTerms;
      // Compute the histogram
      findex->mutex.SemP ();
      index = (findex->index+1)%2;
      findex->mutex.SemV ();
      if (validTerms==1)
      {
	fHisto[index].Compute( modelsArray[0], 0,
			       (float)fAverageDist[0] );
      } else
      {
	cout << "MaxDist " << maxDist << endl;
	fHisto[index].BuildFromMixModels( modelsArray,
					  weights,
					  validTerms,
					  maxDist );
	cout << "Histo built " << endl;
      }
      findex->mutex.SemP ();
      findex->index=index;
      findex->mutex.SemV ();
      delete modelsArray;
      delete weights;

      // Reset the points counter
      fNumPts=0;
      
    } else
    {
      cout << "Error reading data" << endl;
    }
  }
  timeElapsed+=fTimer->Gap();
  return(err);
  
}
  



////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                         protected function members                         //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

//
//    InitProcess
//

inline
ErrorNumber TAdaptEM::InitProcess (int&        oWidth,
				   int&        oHeight,
				   UInt32&     oOffset,
				   UInt8*      &oBase)
{
  TRect           roi;
  UInt32          c[2];
  TBitMapSpec     spec;

  // get source bitmap parameters
  spec      = fbm->GetSpec ();

  if ((spec.encoding >> 8) != 4)
    // can only manage 4 byte per pixel source bitmap
    return kErrCantHandleEncoding;

  oOffset   = spec.lineOffset;

  // defines the ROI
  // wait for mutex
  bb->mutex.SemP ();
  if (bb->roi == TRect(0,0,0,0))
  {
    return kErrSingularity;
  }
  else
  {
    roi = bb->roi;
  }
  bb->mutex.SemV ();

  // size of the area to process
  oWidth  = roi.right   - roi.left;
  oHeight = roi.bottom  - roi.top;
  c[0]=(roi.right + roi.left)/2;
  c[1]=(roi.top + roi.bottom)/2;
  // Source bitmap initialization:
  //  oBase = (UInt8 *)(fbm->GetBase()) + roi.top * oOffset + roi.left * 4;
  oBase = (UInt8 *)(fbm->GetBase()) + c[1] * oOffset + c[0] * 4;
  return kErrNoErr;
}
