/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// UGaussianInv.h
//  Nuria Oliver 1997, nuria@media.mit.edu

#ifndef _UGAUSSIANINV_H_
#define _UGAUSSIANINV_H_

#include "UVector2.h"
#include "UMatrix3.h"
#include "UMatrix4.h"
#include <math.h>

#define  M_PI     3.14159265359

template<class T>
class TGaussianInv
{
private:
    unsigned int m_nSampleCount;
    BOOL		 m_bSingular;
    unsigned int m_nArea; // number of samples that have been used for
    // computing the current statistics
    TVector2<T>  m_vecSampleSum;
    TMatrix3<T>  m_matSampleSumSquares;
    TVector2<T>  m_vecMean;
    TMatrix3<T>  m_matInvCov;
public:
    
    TGaussianInv() 
    { setZero(); }
    TGaussianInv(  int sampleCount, T* sampleData )
    { setZero();set( sampleCount, sampleData); }
    TGaussianInv(  T* mean,  T* cov )
    { setZero();set( mean, cov ); }
    TGaussianInv(  TVector2<T>& mean,  TMatrix3<T> cov )
    { setZero();set( mean, cov ); }
    TGaussianInv( TGaussianInv<T>& g)
    { setZero();set( g ); }
    ~TGaussianInv(){}
	
    // Acces to data members
	T* getMeanData() { return m_vecMean.getData(); }
	TVector2<T> getMean() { return m_vecMean; }
	T  getElemMean( int index=0 ) { return m_vecMean(index); }
	TMatrix3<T> getInvCov() { return m_matInvCov; }
	int getCov(TMatrix3<T>&);
	T* getInvCovData() { return m_matInvCov.getData(); }
	T  getElemInvCov( int index=0 ) { return m_matInvCov(index); }
	TVector2<T> getSampleSum() { return m_vecSampleSum; }
	T* getSampleSumData() { return m_vecSampleSum.getData(); }
	T  getElemSampleSum( int index=0 ) { return m_vecSampleSum(index); }
	TMatrix3<T> getSampleSumSquares() { return m_matSampleSumSquares; }
	T* getSampleSumSquaresData() { return m_matSampleSumSquares.getData(); }
	T  getElemSampleSumSquares( int index=0 ) 
    { return m_matSampleSumSquares(index); }
	unsigned int getSampleCount() { return m_nSampleCount; }
	unsigned int getArea() { return m_nArea; }
	T getDet();
	BOOL getSingular() { return m_bSingular; }
	
	void setMean( T *m ){ m_vecMean.set(m);}
	void setInvCov( T *m ){ m_matInvCov.set( m );}
	void setSampleSum( T *m ){ m_vecSampleSum.set( m );}
	void setSampleSumSquares( T *m ){ m_matSampleSumSquares.set(m); }
	void setSampleCount( unsigned int c ){ m_nSampleCount=c; }
	void setZero();
    
    TGaussianInv<T>& set(  int sampleCount, T* sampleData );
    TGaussianInv<T>& set( T* mean, T* invCov );
    TGaussianInv<T>& set( TVector2<T>& mean, TMatrix3<T>& invCov );
    TGaussianInv<T>& set(  TGaussianInv<T> & );
    
    float getMahalanobisSquare( T p1, T p2 );
    float getMahalanobisSquare( TVector2<T>& v ){return(getMahalanobisSquare(v(0), v(1)));}
    float getProbability( T p1, T p2 );
    float getProbability( TVector2<T>& v ){return(getProbability(v(0), v(1)));}
    float getMahalanobis( T p1, T p2 );
    float getMahalanobis( TVector2<T>& v ){return(getMahalanobis(v(0), v(1)));}
    
    void addSample(  T p1,  T p2 );
    void addSample( TVector2<T>& v){addSample(v(0), v(1));}
    int getGaussianInv();
    // Gets gaussian and eigen vectors/values
    int getGaussianInv(T* resL, T* resS );
	int getGaussianInv( TVector2<T>& resL, TMatrix4<T>& resS )
    { return getGaussianInv( resL.getData(), resS.getData() ); }
    
    int getEigen( T* resL, T* resS );
    int getEigen( TVector2<T>& resL, TMatrix4<T>& resS );
	
    // this functions add two gaussians to get another gaussian
    int setAddEq( TGaussianInv<T>& g, float w = .5 );
    int setAdd( TGaussianInv<T>& g1, TGaussianInv<T>& g2, float w = .5 );
	
    // function for narrowing a gaussian or making it converge to another
    // gaussian. 
    int setMerge( TGaussianInv<T>& g1, TGaussianInv<T>& g2 );
    int setMergeEq( TGaussianInv<T>& g );
	
	// Write to file
	void	Write( CFile	&file );
};

template <class T>
int TGaussianInv<T>::getCov(TMatrix3<T>& m)
{
    //if singular returns error
    if ( m_bSingular == TRUE ) return 1;
    m.set( m_matInvCov );
    m.setInvertEq();
    return 0;
}


template<class T> 
T TGaussianInv<T>::getDet(  )
{
	double det=0;
    if ( m_bSingular == TRUE ) {
		return ((T)0);
    } else {
		det = m_matInvCov.getDet();
		return(T(det));
    }
}

#define MAXEXP         30.
#include <math.h>

/*****************************************SETS*********************************/

template<class T>
void
TGaussianInv<T>::setZero()
{
    m_nSampleCount=m_nArea=0;
    m_bSingular=FALSE;
    m_vecMean.setZero();
    m_matInvCov.setIdentity();
    m_matSampleSumSquares.setZero();
    m_vecSampleSum.setZero();
}

template<class T> 
TGaussianInv<T>&
TGaussianInv<T>::set(int sampleCount, T* sampleData )
{
    TVector2<T> v;
    m_vecMean.setZero();
    m_matInvCov.setZero();
    for ( register int i=0; i<sampleCount; i++ ) {
		v.set( sampleData );
		m_vecMean.setAddEq( v );
		m_matInvCov.setAddOuterProduct( v );
		sampleData +=2;	
    }
    m_vecMean.setDivideEq((float)sampleCount);
    m_matInvCov.setDivideEq((float)sampleCount);
    m_matInvCov.setSubtractOuterProduct( m_vecMean );
    
    if ( !m_matInvCov.setInvertEq() ) {
		m_bSingular=FALSE;
    } else {
		m_bSingular=TRUE;
		m_matInvCov.setIdentity();
    }
    m_nArea = m_nSampleCount;
    return *this;
}

template<class T> 
TGaussianInv<T>& TGaussianInv<T>::set( T* m, T* c )
{
    m_vecMean.set( m );
    m_matInvCov.set( c );
    if ( m_matInvCov.setInvertEq() ) {
		m_bSingular = TRUE;
		m_matInvCov.setIdentity();
    } else {
		m_bSingular = 0;
    }
    return *this;
}

template<class T> 
TGaussianInv<T>& TGaussianInv<T>::set( TVector2<T>& m, TMatrix3<T>& c )
{
    return(set( m.getData(), c.getData() ));
}

template<class T> 
TGaussianInv<T>& TGaussianInv<T>::set(  TGaussianInv<T>& g )
{
    m_vecMean.set( g.getMean() );
    m_matInvCov.set( g.getInvCov() );
    m_bSingular = g.getSingular();
    m_vecSampleSum.set( g.getSampleSum() );
    m_matSampleSumSquares.set( g.getSampleSumSquares() );
    m_nSampleCount = g.m_nSampleCount;
    m_nArea = g.m_nArea;
    return *this;
}

template<class T> 
float TGaussianInv<T>::getMahalanobisSquare(  T p1,  T p2 )
{
    register T temp[2];
    register T dif[2];
    dif[0]=m_vecMean(0)-p1;
    dif[1]=m_vecMean(1)-p2;
    if ( m_bSingular ) {
		return(dif[0]*dif[0]+dif[1]*dif[1]);
    } else {
		temp[0]=m_matInvCov(0)*dif[0]+m_matInvCov(1)*dif[1];
		temp[1]=m_matInvCov(1)*dif[0]+m_matInvCov(2)*dif[1];
		return( dif[0]*temp[0]+dif[1]*temp[1] );
    }
}


template<class T> 
float TGaussianInv<T>::getProbability(  T p1,  T p2 )
{
    register T temp[2];
    register T dif[2];
    register double aux;
    register double det;
    dif[0]=m_vecMean(0)-p1;
    dif[1]=m_vecMean(1)-p2;
    if ( m_bSingular ) {
		return ((T)0);
    } else {
		temp[0]=m_matInvCov(0)*dif[0]+m_matInvCov(1)*dif[1];
		temp[1]=m_matInvCov(1)*dif[0]+m_matInvCov(2)*dif[1];
		aux = (double)(dif[0]*temp[0]+dif[1]*temp[1]);	
		det = m_matInvCov.getDet();
		if ( (aux > MAXEXP) || (det<=0) )
			return(0.0);
		else
			aux = sqrt(det)*exp(-.5*aux)/(2*M_PI);
		return(T(aux));
    }
}


template<class T> 
float TGaussianInv<T>::getMahalanobis(  T p1,  T p2 )
{
    return( (float) sqrt( getMahalanobisSquare( p1, p2 ) ) );
}


template<class T> 
void TGaussianInv<T>::addSample(  T p1,  T p2 )
{
    m_nSampleCount++;
    m_nArea++;
    m_vecSampleSum.setAddEq( p1, p2 );
    m_matSampleSumSquares.setAddOuterProduct( p1, p2 );
}


template<class T> 
int TGaussianInv<T>::getGaussianInv()
{
	
    // if no samples, do nothing
    if ( !m_nSampleCount ) {
		return 0;
    } else {
		m_vecMean.set( m_vecSampleSum );
		m_vecMean/=(float)m_nSampleCount;
		m_matInvCov.set( m_matSampleSumSquares );
		m_matInvCov/=(float)m_nSampleCount;
		m_matInvCov.setSubtractOuterProduct( m_vecMean );
		// this other definition is broken because of use of Vector2
		//m_matInvCov.setSubtractOuterProduct( m_vecMean );
		m_nArea = m_nSampleCount;
		m_vecSampleSum.setZero();
		m_matSampleSumSquares.setZero();
		m_nSampleCount=0;
		m_bSingular=FALSE;
		if ( m_matInvCov.setInvertEq() ) {
			m_bSingular=1;
			m_matInvCov.setIdentity();
		} 
		return( (int)m_bSingular );
    }   
}


template<class T> 
int TGaussianInv<T>::getGaussianInv( T* resL, T* resS )
{
    TMatrix3<T> cov;
    // if no samples, do nothing
    if ( !m_nSampleCount ) {
		return 0;
    } else {
		m_vecMean.set( m_vecSampleSum );
		m_vecMean/=(float)m_nSampleCount;
		m_matInvCov.set( m_matSampleSumSquares );
		m_matInvCov/=(float)m_nSampleCount;
		m_matInvCov.setSubtractOuterProduct( m_vecMean );
		m_nArea = m_nSampleCount;
		m_vecSampleSum.setZero();
		m_matSampleSumSquares.setZero();
		m_nSampleCount=0;
		m_bSingular=0;
		// Save covariance before inverting for
		// computing the eigen vectors/values
		cov.set(m_matInvCov);
		if ( m_matInvCov.setInvertEq() ) {
			m_bSingular=TRUE;
			resL[0]=resL[1]=0.0;
			resS[0]=resS[1]=resS[2]=resS[3]=0.0;
			m_matInvCov.setIdentity();
		} else {
			eigen( cov, resL, resS );
		}
		return( (int)m_bSingular );
    }   
}

//The resulting eigenvectors matrix has each eigenvector in each COLUMN!!
//         resS[0]=eigVec0x    resS[1]=eigVec1x
//  resS=
//         resS[2]=eigVec0y    resS[3]=eigVec1y

template<class T>
int eigen( TMatrix3<T>& cov, T* resL, T* resS )
{
    T a = cov(0);
    T b = cov(1);
    T c = cov(1);
    T d = cov(2);
	
	
    // Compute eigenvalues, solve quadratic equation
    // ------------------------------------------------
    double B = -(a+d);
    double C = a*d - b*c;
	
    double rad = sqrt((float)(B*B) - (float)(4.*C));
    double l0 = (-B - rad) / 2.;
    double l1 = (-B + rad) / 2.;
	
    // Compute eigenvectors
    // ------------------------------------------------
    //  [ a-l   b  ][ x1 ]
    //  [  c   d-l ][ x2 ] is zero
    // 1. find the most well-conditioned row (2 evalues, 2 rows each)
    double row[4][2];
    double norm[4];
    int best = 0;
    int i;
	
    row[0][0] = a - l0;
    row[0][1] = b;
    row[1][0] = c;
    row[1][1] = d - l0;
    row[2][0] = a - l1;
    row[2][1] = b;
    row[3][0] = c;
    row[3][1] = d - l1;
	
    norm[0] = row[0][0]*row[0][0] + row[0][1]*row[0][1];
    norm[1] = row[1][0]*row[1][0] + row[1][1]*row[1][1];
    norm[2] = row[2][0]*row[2][0] + row[2][1]*row[2][1];
    norm[3] = row[3][0]*row[3][0] + row[3][1]*row[3][1];
	
    best = 0;
    for(i=1; i<4; i++)
    {
		if (norm[i] > norm[best]) 
			best = i;
    }
    // 2. now check that this norm is good enough precision..    
    if ( norm[best] < 1e-10 )
	{
		// ...if not, we have a circular distribution
		resL[0] = resL[1] = T(1.0);
		resS[0] = resS[3] = T(1.0);
		resS[1] = resS[2] = T(0.0);
		return 1;
	}
    // ...if so, ...
    // 3. normalize the best row and find the eigenvectors
    {
		int one,zero;
		if(best < 2) 
			one = 0, zero = 1;
		else
			one = 1, zero = 0;
		
		
		norm[best] = sqrt(norm[best]);
		
		resS[0+zero] =  T(row[best][0] / norm[best]);
		resS[2+zero] =  T(row[best][1] / norm[best]);
		resS[0+one] =  T(resS[2+zero]);
		resS[2+one] = T(-resS[0+zero]);
    }
    
    resL[0] = T(l0);
    resL[1] = T(l1);
    // Transpose the eigenvectors matrix
    T tmp = resS[1];
    resS[1] = resS[2];
    resS[2] = tmp;
	
    // We force the first eigenvector to have a positive first component
    if ( resS[0] < 0 ) {
		resS[0]=-resS[0];
		resS[1]=-resS[1];
    }
    if ( resS[3] < 0 )
    {
		resS[2]=-resS[2];
		resS[3]=-resS[3];
    }
    /*if (fabsf(l0) < fabsf(l1))
    {
	
	  tmp = resL[0];
	  resL[0] = resL[1];
	  resL[1] = tmp;
	  
		tmp = resS[0];
		resS[0] = resS[1];
		resS[1] = tmp;
		
		  tmp = resS[2];
		  resS[2] = resS[3];
		  resS[3] = tmp;
}*/
	
    return 0;
}


template<class T> 
int TGaussianInv<T>::getEigen( TVector2<T>& resL, TMatrix4<T>& resS )
{
    int res=0;
    // if m_bSingular returns error and results set to 0
    if ( m_bSingular ) {
		resL.setZero();
		resS.setZero();
		res=1;
    } else {
		// needs to invert the inverse of the covariance
		TMatrix3<T> aux;
		m_matInvCov.getInvert(aux);
		res = eigen( aux, resL.getData(), resS.getData() );
    }
    return res;
}



template<class T> 
int TGaussianInv<T>::getEigen( T* resL, T* resS )
{
    TMatrix3<T> aux;
	m_matInvCov.getInvert(aux);
    return eigen( aux, resL, resS );
}

template<class T> 
int TGaussianInv<T>::setAdd(	TGaussianInv<T>& g1, 
							TGaussianInv<T>& g2, float w )
{
    //Adds g1 and g2 weighted by w and (1-w) and saves the result in this
    //if any singular, return error
    if ( (g1.getSingular() == TRUE) || (g2.getSingular() == TRUE) )
		return 1;
    m_nArea = g1.getArea()+g2.getArea();
    TMatrix3<T> sqr1;
    TMatrix3<T> sqr2;
    TMatrix3<T> m1m1;
    TMatrix3<T> m2m2;
    TMatrix4<T> m1m2;
    TMatrix4<T> m2m1;
    TVector2<T> m1( g1.getMean() );
    TVector2<T> m2( g2.getMean() );
    g1.getCov(sqr1);
    g2.getCov(sqr2);
    m1m1.setOuterProduct( m1 );
    m2m2.setOuterProduct( m2 );
    m1m2.setOuterProduct( m1, m2 );
    m2m1.setOuterProduct( m2, m1 );
    sqr1.setAddEq( m1m1 );
    sqr2.setAddEq( m2m2 );
    m_vecMean.set( m1 );
    m_vecMean.setScaleEq( w );
    m2.setScaleEq( 1-w );
    m_vecMean.setAddEq( m2 );
    sqr1.setScaleEq( w );
    sqr2.setScaleEq( 1-w );
    m1m1.setScaleEq( -w*w );
    m2m2.setScaleEq( -(1-w)*(1-w) );
    m1m2.setScaleEq( -w*(1-w) );
    m2m1.setScaleEq( -(1-w)*w );
    m_matInvCov.set( sqr1 );
    m_matInvCov.setAddEq( sqr2 );
    m_matInvCov.setAddEq( m1m1 );
    m_matInvCov.setAddEq( m2m2 );
    TMatrix3<T> m1m2Sim( m1m2(0)*2, m1m2(1)+m2m1(1), 2*m1m2(3) );
    m_matInvCov.setAddEq( m1m2Sim );
    m_bSingular = (m_matInvCov.setInvertEq() ? TRUE : FALSE);
    return 0;
}


template<class T> 
int TGaussianInv<T>::setAddEq( TGaussianInv<T>& g, float w )
{
    //Adds g1 and g2 weighted by w and (1-w) and saves the result in this
    //if any singular, return error
    return setAdd( g, *this, w );
}


template<class T> 
int TGaussianInv<T>::setMerge(	TGaussianInv<T>& g1, 
							  TGaussianInv<T>& g2 )
{
    // the formulas for approaching one gaussian to another are:
    //   invCTot = invC1 + invC2
    //   meanTot = CTot*(invC1*m1+invC2*m2)
	
    if ( (g1.getSingular() == TRUE) || (g2.getSingular()==TRUE) ) return 1;
    set( g1 );
    TMatrix3<T> oldInvCov( getInvCov() );
    TMatrix3<T> cov;
    TMatrix3<T> invCov2( g2.getInvCov());
    m_matInvCov.setAddEq( g2.getInvCov() );
    TVector2<T> aux( g2.getMean() );
    // should be invertible
    getCov(cov);
    m_vecMean.setMultiplyEq( oldInvCov );
    aux.setMultiplyEq( invCov2 );
    m_vecMean.setAddEq( aux );
    m_vecMean.setMultiplyEq( cov );
    m_nArea = g1.getArea()+g2.getArea();
    return 0;
}

template<class T> 
int TGaussianInv<T>::setMergeEq( TGaussianInv<T>& g )
{
    // the formulas for approaching one gaussian to another are:
    //   invCTot = invC1 + invC2
    //   meanTot = CTot*(invC1*m1+invC2*m2)
	if ( (m_bSingular==TRUE) || (g.getSingular()==TRUE) ) return 1;
	TMatrix3<T> oldInvCov( getInvCov() );
	TMatrix3<T> cov;
	TMatrix3<T> invCov2( g.getInvCov());
	m_matInvCov.setAddEq( g.getInvCov() );
	TVector2<T> aux( g.getMean() );
	// should be invertible
	getCov(cov);
	m_vecMean.setMultiplyEq( oldInvCov );
	aux.setMultiplyEq( invCov2 );
	m_vecMean.setAddEq( aux );
	m_vecMean.setMultiplyEq( cov ); 
	m_nArea+=g.getArea();
	return 0;  
}


template<class T> 
void TGaussianInv<T>::Write( CFile &f )
{
	m_vecMean.Write( f );
	m_matInvCov.Write( f );
}

#endif /* _UGAUSSIANINV_H_ */
