/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// File       : UKalman.h
// Creation   : 07/1996
// Goal       : declares the TKalman class, a class that smooth estimation of
//              a 2 dimensional vector over time
//              Based on F. Kirouches algorithm.
//
// Tab        : no
//
//  Nuria Oliver 1997, nuria@media.mit.edu

#ifndef _UKALMAN_H
#define _UKALMAN_H

// project includes
#include "UErrors.h"

static const float  kKalDefTimeDev[2] = { 0, 0 };
static const float  kKalDefObsCov[4]  = { 1, 0.5, 0.5, 1 };


static const int    kKalNbVal   = 26;
// nb of value in state array

static const int    kKalVecInd  = 0;
// index or vector values (2 values)

static const int    kKalCovInd  = 2;
// index of covariance values (4 values)

static const int    kKalInvInd  = 6;
// index of the inverse of the covariance values (4 values)

static const int    kKalTimInd  = 10;
// index of time deviation values (2 values)

static const int    kKalSavInd  = 12;
// index of duplication space (6 values)

static const int    kKalObsInd  = 18;
// index of observation covariance matrix (4 values)

static const int    kKalObIInd  = 22;
// index of observation inverse covariance matrix (4 values)


class TTimer;
class TKalman
{
public:
	
    ErrorNumber       Init ();
    // initialise the Kalman filter
	
    void              Dispose ();
    // disposes the Kalman filter
	
    ErrorNumber       Update (float* obs);
    // integrate the observation <obs> that is an array of 2 values
	
    inline
		ErrorNumber       Update (float* obs, float* state);
    // same as previous "update" except that the new state is returned
    //  in <state>. <obs> and <state> may be identical.
	
    inline
		void              Reset ();
    // reset the estimation. The next observation will define the estimation
	
    inline
		float*            GetState ();
    // returns a pointer to the internal state of the filter
	
    //inline
    ErrorNumber       SetObsCovariance (float* cov);
    // set the observation covariance matrix (4 values)
    // returns an error if <obs> is not invertible
	
    //inline
    void              GetObsCovariance (float* cov);
    // get the observation covariance matrix (4 values)
	
    //inline
    void              SetTimeDeviation (float* timeDeviation);
    // set the time deviation parameters (2 values)
	
    //inline
    void              GetTimeDeviation (float* timeDeviation);
    // get the time deviation parameters (2 values)
	
    //inline
    ErrorNumber       Invert (float* source, float* dest);
    // invert the <source> matrix in <dest>
    // <source and <dest> may be indentical
    // returns an error if <source> is not invertible
	
    //inline
    void              AddMat (float* a, float* b, float* dest);
    // add the two matrix <a> and <b> and put the result in <dest>
    // <dest> may be <a> or <b>
	
    //inline
    void              AddVect (float* a, float* b, float* dest);
    // add the two vectors <a> and <b> and put the result in <dest>
    // <dest> may be <b>
	
    //inline
    void              MultVect (float* a, float* b, float* dest);
    // multiplicate matrix <a> anv vector <b> and put the result in <dest>
    // <dest> may be <a> or <b>
	
	
	
	
protected:
	
    TTimer           *m_pTimer;             // measure time between computations
    float             m_fState[kKalNbVal];  // stores current state
    BOOL              m_bInitialised;       // is the filter initialised ?
	
    //inline
    void              ApplyTimeDeviation ();
	
    //inline
    void              CopyState (float* src, float* dst);
};






////////////////////////////////////////////////////////////////////////////////
//                                                                             //
//                      inline function members definition                    //
//                                                                             //
////////////////////////////////////////////////////////////////////////////////

//
//    Reset
//

void TKalman::Reset ()
{
	m_bInitialised = FALSE;
}




//
//    GetState
//

inline
float* TKalman::GetState ()
{
	return m_fState;
}




//
//    Update
//

inline
ErrorNumber TKalman::Update (float* obs, float* state)
{
	ErrorNumber   err;
	
	err = this->Update (obs);
	state[0] = m_fState[kKalVecInd];
	state[1] = m_fState[kKalVecInd+1];
	
	return err;
}




#endif
